#!/usr/bin/env python

from argparse import ArgumentParser
import yaml
import warnings
import multiprocessing
import subprocess
import shlex
import os
import sys
import re

from multiprocessing.pool import ThreadPool


class RadishOptions:
    def __init__(self):

        # default values
        self.res_dictionary = {'NLL': 1, 'NNLL': 2, 'N3LL': 3}
        self.order_dictionary = {'LO': 1, 'NLO': 2, 'NNLO': 3}
        self.matched = False

        # run parameters
        self.process = "DY"
        self.observable = "ptB"
        self.run_name = "radish_example"
        self.output_folder = None
        self.onshell = False
        self.fixedMY = False
        self.fixed_M = 91
        self.fixed_Y = 0

        # event generation parameters
        self.seed = None
        self.born_events = 1000
        self.shower_events = 1000

        # collider parameters
        self.beams = "pp"
        self.Ecm = 13000
        self.pdf_set = "NNPDF31_nnlo_as_0118"
        self.pdf_member = 0

        # resummation parameters
        self.res_order = 3
        self.dynamical = True
        self.p = 4
        self.differential = True
        self.prime = False

        # jetveto parameters
        self.R = 0.4
        self.JV = 30

        # matching parameters
        self.fixed_order = 3
        self.scheme = "multa"
        self.resum_only = True
        self.normalise = False

        # scales parameters
        self.res_scale_vary = 2
        self.ren_scale_vary = 2
        self.fact_scale_vary = 2

        # user cuts
        self.nocuts = False
        self.mmin = 10
        self.mmax = 13000
        self.ymin = -99
        self.ymax = 99

        # extra options
        self.vmin = 1
        self.vmax = 300
        self.nbins = 300
        self.binscale = 'lin'

    def set_run_options(self, dict):
        if 'process' in dict:
            self.process = dict.get("process")
        if 'observable' in dict:
            self.observable = dict.get("observable")
        if 'run_name' in dict:
            self.run_name = dict.get("run_name")
        if 'output_folder' in dict:
            self.output_folder = dict.get("output_folder")
        if 'onshell' in dict:
            self.onshell = dict.get("onshell")
        if (self.onshell and not ((self.process == 'H') or (self.process == 'DY'))):
            raise ValueError('Inclusive flag only possible for process H, DY.')
        if 'fixed_MY' in dict:
            self.fixedMY = dict.get("fixed_MY")
        if 'fixed_M' in dict:
            self.fixed_M = dict.get("fixed_M")
        if 'fixed_Y' in dict:
            self.fixed_Y = dict.get("fixed_Y")

    def set_generation_options(self, dict):
        if 'seed' in dict:
            self.seed = dict.get("seed")
        if 'born_events' in dict:
            self.born_events = dict.get("born_events")
        if 'shower_events' in dict:
            self.shower_events = dict.get("shower_events")

    def set_collider_options(self, dict):
        if 'beams' in dict:
            self.beams = dict.get("beams")
        if 'Ecm' in dict:
            self.Ecm = dict.get("Ecm")
        if 'pdf' in dict:
            if('set' in dict.get("pdf")):
                self.pdf_set = dict.get("pdf").get('set')
            if('member' in dict.get("pdf")):
                self.pdf_member = dict.get("pdf").get("member")

    def set_resummation_options(self, dict):
        if 'order' in dict:
            self.res_order = self.res_dictionary.get(dict.get("order"))
            if (self.res_order == None):
                raise ValueError('Resummation order can be NLL, NNLL, or N3LL')
        if 'modified_logs_exp' in dict:
            self.p = dict.get("modified_logs_exp")
        if 'differential' in dict:
            self.differential = dict.get("differential")
        if 'prime' in dict:
            self.prime = dict.get("prime")
        if 'jetveto' in dict:
            if 'jet_radius' in dict.get("jetveto"):
                self.R = dict.get("jetveto").get("jet_radius")
            if 'jet_veto' in dict.get("jetveto"):
                self.JV = dict.get("jetveto").get("jet_veto")

    def set_matching_options(self, dict):
        if 'order' in dict:
            self.fixed_order = self.order_dictionary.get(dict.get("order"))
            if (self.fixed_order == None):
                raise ValueError('Matching order can be LO, NLO, or NNLO')
        if 'scheme' in dict:
            self.scheme = dict.get("scheme")
        if 'normalise' in dict:
            self.normalise = dict.get("normalise")
            if ((dict.get("normalise") == True) and (dict.get("fixed_order_file") == 'none')):
                warnings.warn(
                    'No fixed order file present. Result will be normalised to Born.')

    def set_cut_options(self, dict):
        if 'nocuts' in dict:
            self.nocuts = dict.get('nocuts')
        if 'boson_mass' in dict:
            if 'min' in dict.get("boson_mass"):
                self.mmin = dict.get("boson_mass").get("min")
            if 'max' in dict.get("boson_mass"):
                self.mmax = dict.get("boson_mass").get("max")
        if 'boson_rapidity' in dict:
            if 'min' in dict.get("boson_rapidity"):
                self.ymin = dict.get("boson_rapidity").get("min")
            if 'max' in dict.get("boson_rapidity"):
                self.ymax = dict.get("boson_rapidity").get("max")

    def set_binning_options(self, dict):
        if 'obsmin' in dict:
            self.vmin = dict.get('obsmin')
        if 'obsmax' in dict:
            self.vmax = dict.get('obsmax')
        if 'nbins' in dict:
            self.nbins = dict.get('nbins')
        if 'binscale' in dict:
            self.binscale = dict.get('binscale')

    def construct_output_file(self, scale):
        fixed_order = ""
        if (self.matched):
            fixed_order = "+" + \
                str(self.inverse_mapping(self.order_dictionary).get(self.fixed_order))
        res_order = str(self.inverse_mapping(
            self.res_dictionary).get(self.res_order))
        if(self.prime):
            res_order = res_order + "p"
        order = res_order + fixed_order + "_"
        suffix = ".diff"
        if (not self.differential):
            suffix = ".cum"
        if (not self.output_folder is None):
            os.makedirs(str(self.output_folder), exist_ok=True)
            return str(self.output_folder)+"/"+str(self.run_name)+"_"+order+scale+suffix
        else:
            return str(self.run_name)+"_"+scale+suffix

    def arguments(self):
        argument = ""
        argument += " -collider " + str(self.beams)
        argument += " -proc " + str(self.process)
        argument += " -observable " + str(self.observable)
        argument += " -sqrts " + str(self.Ecm)
        argument += " -pdf_name " + str(self.pdf_set)
        argument += " -pdf_member " + str(self.pdf_member)
        if(not self.onshell):
            argument += " -ncall1 " + str(self.born_events)
            argument += " -ncall2 " + str(self.born_events)
            if(self.nocuts):
                argument += " -nocuts"
                argument += " -mmin " + str(self.mmin)
                argument += " -mmax " + str(self.mmax)
            if(not self.nocuts):
                argument += " -mmin " + str(self.mmin)
                argument += " -mmax " + str(self.mmax)
                argument += " -etamin " + str(self.ymin)
                argument += " -etamax " + str(self.ymax)
            if(self.dynamical):
                argument += " -dynamical "
        argument += " -iseed " + str(self.seed)
        argument += " -nev " + str(self.shower_events)
        argument += " -order " + str(self.res_order)
        argument += " -p " + str(self.p)
        argument += " -fixed_order " + str(self.fixed_order)
        argument += " -scheme " + str(self.scheme)
        argument += " -R " + str(self.R)
        argument += " -JV " + str(self.JV)
        if(self.differential):
            argument += " -differential"
        if(self.prime):
            argument += " -prime"
        if(not self.normalise):
            argument += " -cross-section"
        if(self.fixedMY):
            argument += " -singleYMpoint"
            argument += " -Mfixed " + str(self.fixed_M)
            argument += " -Yfixed " + str(self.fixed_Y)
        return argument

    def binarguments(self):
        argument = ""
        argument += " -vmin " + str(self.vmin)
        argument += " -vmax " + str(self.vmax)
        argument += " -nbins " + str(self.nbins)
        if (self.binscale == 'lin'):
            argument += ""
        elif (self.binscale == 'log'):
            argument += " -log "
        elif (self.binscale == 'log10'):
            argument += " -log10 "
        else:
            raise ValueError(
                "Unknown option for binscale. Available options are lin, log, log10.")
        return argument

    def inverse_mapping(self, f):
        return f.__class__(map(reversed, f.items()))


def submit_job(cmd):
    """ This runs in a separate thread. """
    # subprocess.call(shlex.split(cmd))  # This will block until cmd finishes
    p = subprocess.Popen(shlex.split(
        cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    return (out, err)


def print_banner():
    print("")
    print(" ___________________________________________________________________________ ")
    print("| Version: 3.0.0                                                  June 2020 |")
    print("|                                                                           |")
    print("|\033[1;92m"+"                      ____           _ ___ ____  _   _                     "+"\033[0m|")
    print("|\033[1;92m"+"                     |  _ \ __ _  __| |_ _/ ___|| | | |                    "+"\033[0m|")
    print("|\033[1;92m"+"                     | |_) / _` |/ _` || |\___ \| |_| |                    "+"\033[0m|")
    print("|\033[1;92m"+"                     |  _ < (_| | (_| || | ___) |  _  |                    "+"\033[0m|")
    print("|\033[1;92m"+"                     |_| \_\__,_|\__,_|___|____/|_| |_|                    "+"\033[0m|")
    print("|                                                                           |")
    print("|                    \033[1mRad\033[0miation off \033[1mI\033[0mnitial \033[1mS\033[0mtate \033[1mH\033[0madrons                    |")
    print("|       Written by W. Bizon, P. Monni, E. Re, L. Rottoli, P. Torrielli      |")
    print("|                                                                           |")
    print("|                              Developers:                                  |")
    print("|                   P. Monni      (pier.monni@cern.ch)                      |")
    print("|                   E. Re         (emanuele.re@lapth.cnrs.fr)               |")
    print("|                   L. Rottoli    (luca.rottoli@cern.ch)                    |")
    print("|                   P. Torrielli  (torriell@to.infn.it)                     |")
    print("|                                                                           |")
    print("|                Available under the GNU public license                     |")
    print("|                                                                           |")
    print("| If you use it or any derivative of it in scientific work you should cite  |")
    print("| P. Monni, E. Re, P. Torrielli: Phys.Rev.Lett. 116 (2016) no.24, 242001    |")
    print("| W. Bizon, P. Monni, E. Re, L. Rottoli, P. Torrielli: JHEP 1802 (2018) 108 |")
    print("|                                                                           |")
    print("| If you use it for the joint resummation of pT and pTj, please cite also   |")
    print("| P. Monni, L. Rottoli, P. Torrielli, arXiv:1909.04704                      |")
    print("|                                                                           |")
    print("|                                                                           |")
    print(" ---------------------------------------------------------------------------")
    print("")


def main():
    """
    """

    print_banner()

    # read arguments from command line
    parser = ArgumentParser(description='input file')
    parser.add_argument('-i', '--input', dest="infilename",
                        help="input file (yaml format)",
                        metavar="INFILE",
                        required=True)
    parser.add_argument('-n', '--num-runs', dest="max_processes",
                        help="number of processes", type=int,
                        metavar="NPROC", default=1)
    parser.add_argument('-d', '--dry-run', dest="dryrun",
                        action="store_true")
    args = parser.parse_args()

    with open(args.infilename, 'r') as stream:
        try:
            dictionary = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)

    # run settings variables

    options = RadishOptions()

    if 'run' in dictionary:
        options.set_run_options(dictionary.get('run'))

    if 'generation' in dictionary:
        options.set_generation_options(dictionary.get('generation'))

    if 'collider' in dictionary:
        options.set_collider_options(dictionary.get('collider'))

    if 'resummation' in dictionary:
        options.set_resummation_options(dictionary.get('resummation'))

    if 'matching' in dictionary:
        options.set_matching_options(dictionary.get('matching'))

    if 'cuts' in dictionary:
        options.set_cut_options(dictionary.get('cuts'))

    if 'binning' in dictionary:
        options.set_binning_options(dictionary.get('binning'))

    scale_variations = []
    scale_var_string = []
    if 'scales' in dictionary:
        if (dictionary.get('scales').get('scale_settings') == "fixed"):
            options.dynamical = False
        if (dictionary.get('scales').get('res_scale_vary')):
            options.res_scale_vary = dictionary.get(
                'scales').get('res_scale_vary')
        if (dictionary.get('scales').get('ren_scale_vary')):
            options.ren_scale_vary = dictionary.get(
                'scales').get('ren_scale_vary')
        if (dictionary.get('scales').get('fact_scale_vary')):
            options.fact_scale_vary = dictionary.get(
                'scales').get('fact_scale_vary')
        if (dictionary.get('scales').get('variations') == 'none'):
            if (options.dynamical == False):
                if ('xmuF' in dictionary.get('scales')) or ('xmuR' in dictionary.get('scales')
                                                            or (not 'muF' in dictionary.get('scales'))
                                                            or (not 'muR' in dictionary.get('scales'))):
                    raise ValueError(
                        'You are using fixed scales. Please specify the value of factorisation and renormalisation scales')
                elif ('xQ' in dictionary.get('scales') or (not 'Q' in dictionary.get('scales'))):
                    raise ValueError(
                        'You are using fixed scales. Please specify the value of resummation scale.')
                else:
                    Q = dictionary.get('scales').get('Q')
                    muF = dictionary.get('scales').get('muF')
                    muR = dictionary.get('scales').get('muR')
                    scale_variations.append([muF, muR, Q])
                    scale_var_string.append("xmuF10_xmuR10_xQcentral")
            else:
                if ('muF' in dictionary.get('scales')) or ('muR' in dictionary.get('scales')):
                    raise ValueError(
                        'You are using dynamical scales. Please use xmuF and xmuR to specify the values of the factorisarton and renormalisation scales.')
                elif ('Q' in dictionary.get('scales')):
                    raise ValueError(
                        'You are using dynamical scales. Please use xQ to specify the value of the resummation scale.')
                else:
                    xQ = 0.5
                    xmuF = 1
                    xmuR = 1
                    if 'xQ' in dictionary.get('scales'):
                        xQ = dictionary.get('scales').get('xQ')
                    if 'xmuF' in dictionary.get('scales'):
                        xmuF = dictionary.get('scales').get('xmuF')
                    if 'xmuR' in dictionary.get('scales'):
                        xmuF = dictionary.get('scales').get('xmuR')
                    scale_variations.append([xmuF, xmuR, xQ])
                    scale_var_string.append("xmuF10_xmuR10_xQcentral")
        elif (dictionary.get('scales').get('variations') == 'correlated'):
            if (options.dynamical == False):
                if ('xmuF' in dictionary.get('scales')) or ('xmuR' in dictionary.get('scales')
                                                            or (not 'muF' in dictionary.get('scales'))
                                                            or (not 'muR' in dictionary.get('scales'))):
                    raise ValueError(
                        'You are using fixed scales. Please specify the central value of factorisation and renormalisation scales to compute scale variations')
                elif ('xQ' in dictionary.get('scales') or (not 'muF' in dictionary.get('scales'))):
                    raise ValueError(
                        'You are using fixed scales. Please specify the central value of resummation scale to compute scale variations.')
                else:
                    Q = dictionary.get('scales').get('Q')
                    muF = dictionary.get('scales').get('muF')
                    muR = dictionary.get('scales').get('muR')
                    scale_variations.append([muF, muR, Q])
                    scale_variations.append(
                        [options.fact_scale_vary*muF, options.ren_scale_vary*muR, Q])
                    scale_variations.append(
                        [1/options.fact_scale_vary*muF, 1/options.ren_scale_vary*muR, Q])
                    scale_variations.append(
                        [muF, muR, options.res_scale_vary*Q])
                    scale_variations.append(
                        [muF, muR, 1/options.res_scale_vary*Q])
                    scale_var_string.append(
                        "xmuFcentral_xmuRcentral_xQcentral")
                    scale_var_string.append("xmuFup_xmuRup_xQcentral")
                    scale_var_string.append("xmuFdown_xmuRdown_xQcentral")
                    scale_var_string.append("xmuFcentral_xmuRcentral_xQup")
                    scale_var_string.append("xmuFcentral_xmuRcentral_xQdown")
            else:
                if ('muF' in dictionary.get('scales')) or ('muR' in dictionary.get('scales')):
                    raise ValueError(
                        'You are using dynamical scales. Please use xmuF and xmuR to specify the central values of the factorisation and renormalisation scales.')
                elif ('Q' in dictionary.get('scales')):
                    raise ValueError(
                        'You are using dynamical scales. Please use xQ to specify the central value of the resummation scale.')
                else:
                    xQ = 0.5
                    xmuF = 1
                    xmuR = 1
                    if 'xQ' in dictionary.get('scales'):
                        xQ = dictionary.get('scales').get('xQ')
                    if 'xmuF' in dictionary.get('scales'):
                        xmuF = dictionary.get('scales').get('xmuF')
                    if 'xmuR' in dictionary.get('scales'):
                        xmuR = dictionary.get('scales').get('xmuR')
                    scale_variations.append([xmuF, xmuR, xQ])
                    scale_variations.append(
                        [options.fact_scale_vary*xmuF, options.ren_scale_vary*xmuR, xQ])
                    scale_variations.append(
                        [1/options.fact_scale_vary*xmuF, 1/options.ren_scale_vary*xmuR, xQ])
                    scale_variations.append(
                        [xmuF, xmuR, options.res_scale_vary*xQ])
                    scale_variations.append(
                        [xmuF, xmuR, 1/options.res_scale_vary*xQ])
                    scale_var_string.append(
                        "xmuFcentral_xmuRcentral_xQcentral")
                    scale_var_string.append("xmuFup_xmuRup_xQcentral")
                    scale_var_string.append("xmuFdown_xmuRdown_xQcentral")
                    scale_var_string.append("xmuFcentral_xmuRcentral_xQup")
                    scale_var_string.append("xmuFcentral_xmuRcentral_xQdown")
        elif (dictionary.get('scales').get('variations') == 'all'):
            if (options.dynamical == False):
                if ('xmuF' in dictionary.get('scales')) or ('xmuR' in dictionary.get('scales')
                                                            or (not 'muF' in dictionary.get('scales'))
                                                            or (not 'muR' in dictionary.get('scales'))):
                    raise ValueError(
                        'You are using fixed scales. Please specify the central value of factorisation and renormalisation scales to compute scale variations')
                elif ('xQ' in dictionary.get('scales') or (not 'muF' in dictionary.get('scales'))):
                    raise ValueError(
                        'You are using fixed scales. Please specify the central value of resummation scale to compute scale variations.')
                else:
                    Q = dictionary.get('scales').get('Q')
                    muF = dictionary.get('scales').get('muF')
                    muR = dictionary.get('scales').get('muR')
                    scale_variations.append([muF, muR, Q])
                    scale_variations.append(
                        [options.fact_scale_vary*muF, options.ren_scale_vary*muR, Q])
                    scale_variations.append(
                        [1/options.fact_scale_vary*muF, 1/options.ren_scale_vary*muR, Q])
                    scale_variations.append(
                        [options.fact_scale_vary*muF, muR, Q])
                    scale_variations.append(
                        [muF, options.ren_scale_vary*muR, Q])
                    scale_variations.append(
                        [1/options.fact_scale_vary*muF, muR, Q])
                    scale_variations.append(
                        [muF, 1/options.ren_scale_vary*muR, Q])
                    scale_variations.append(
                        [muF, muR, options.res_scale_vary*Q])
                    scale_variations.append(
                        [muF, muR, 1/options.res_scale_vary*Q])
                    scale_var_string.append(
                        "xmuFcentral_xmuRcentral_xQcentral")
                    scale_var_string.append("xmuFup_xmuRup_xQcentral")
                    scale_var_string.append("xmuFdown_xmuRdown_xQcentral")
                    scale_var_string.append("xmuFup_xmuRcentral_xQcentral")
                    scale_var_string.append("xmuFcentral_xmuRup_xQcentral")
                    scale_var_string.append("xmuFdown_xmuRcentral_xQcentral")
                    scale_var_string.append("xmuFcentral_xmuRdown_xQcentral")
                    scale_var_string.append("xmuFcentral_xmuRcentral_xQup")
                    scale_var_string.append("xmuFcentral_xmuRcentral_xQdown")

            else:
                if ('muF' in dictionary.get('scales')) or ('muR' in dictionary.get('scales')):
                    raise ValueError(
                        'You are using dynamical scales. Please use xmuF and xmuR to specify the central values of the factorisation and renormalisation scales.')
                elif ('Q' in dictionary.get('scales')):
                    raise ValueError(
                        'You are using dynamical scales. Please use xQ to specify the central value of the resummation scale.')
                else:
                    xQ = 0.5
                    xmuF = 1
                    xmuR = 1
                    if 'xQ' in dictionary.get('scales'):
                        xQ = dictionary.get('scales').get('xQ')
                    if 'xmuF' in dictionary.get('scales'):
                        xmuF = dictionary.get('scales').get('xmuF')
                    if 'xmuR' in dictionary.get('scales'):
                        xmuR = dictionary.get('scales').get('xmuR')
                    scale_variations.append([xmuF, xmuR, xQ])
                    scale_variations.append(
                        [options.fact_scale_vary*xmuF, options.ren_scale_vary*xmuR, xQ])
                    scale_variations.append(
                        [1./options.fact_scale_vary*xmuF, 1./options.ren_scale_vary*xmuR, xQ])
                    scale_variations.append(
                        [options.fact_scale_vary*xmuF, xmuR, xQ])
                    scale_variations.append(
                        [xmuF, options.ren_scale_vary*xmuR, xQ])
                    scale_variations.append(
                        [1./options.fact_scale_vary*xmuF, xmuR, xQ])
                    scale_variations.append(
                        [xmuF, 1./options.ren_scale_vary*xmuR, xQ])
                    scale_variations.append(
                        [xmuF, xmuR, options.res_scale_vary*xQ])
                    scale_variations.append(
                        [xmuF, xmuR, 1./options.res_scale_vary*xQ])
                    scale_var_string.append(
                        "xmuFcentral_xmuRcentral_xQcentral")
                    scale_var_string.append("xmuFup_xmuRup_xQcentral")
                    scale_var_string.append("xmuFdown_xmuRdown_xQcentral")
                    scale_var_string.append("xmuFup_xmuRcentral_xQcentral")
                    scale_var_string.append("xmuFcentral_xmuRup_xQcentral")
                    scale_var_string.append("xmuFdown_xmuRcentral_xQcentral")
                    scale_var_string.append("xmuFcentral_xmuRdown_xQcentral")
                    scale_var_string.append("xmuFcentral_xmuRcentral_xQup")
                    scale_var_string.append("xmuFcentral_xmuRcentral_xQdown")
        else:
            raise ValueError(
                'Possible options for scale_settings are: [none, correlated, all] ')

    fixed_order_files = []
    extra_argument = " -resum-only " + options.binarguments()
    if 'fixed_order_file' in dictionary.get('matching'):
        if (not dictionary.get('matching').get('fixed_order_file') == 'none'):
            options.matched = True
            extra_argument = ""
            fixed_order_files = dictionary.get(
                'matching').get('fixed_order_file')
            if (dictionary.get('scales').get('variations') == 'none'):
                nscales = len(scale_variations)
            else:
                nscales = len(scale_variations)-2
            if(not len(dictionary.get('matching').get('fixed_order_file')) == nscales):
                raise ValueError("Number of fixed order files: " +
                                 str(len(dictionary.get('matching').get(
                                     'fixed_order_file')))
                                 + " different from number of scales: " + str(len(scale_variations)))
            for j in range(2):
                fixed_order_files.append(fixed_order_files[0])

    command = ${PYTHON_BIN}

    max_processes = args.max_processes

    pool = ThreadPool(min(multiprocessing.cpu_count(), max_processes))
    results = []
    base_arguments = extra_argument + options.arguments()
    for i in range(len(scale_variations)):
        arguments = base_arguments
        if(options.dynamical == False):
            scale_arguments = " -muF " + str(scale_variations[i][0]) + " -muR " + str(
                scale_variations[i][1]) + " -Q " + str(scale_variations[i][2])
        else:
            scale_arguments = " -xmuF " + str(scale_variations[i][0]) + " -xmuR " + str(
                scale_variations[i][1]) + " -xQ " + str(scale_variations[i][2])
        arguments += scale_arguments
        outputfile = options.construct_output_file(scale_var_string[i])
        arguments += " -out " + outputfile
        if (options.onshell):
            arguments += " -radish_inclusive "
        if(len(fixed_order_files) != 0):
            arguments += " -in " + fixed_order_files[i]
        if(args.dryrun):
            print(command+arguments)
        else:
            results.append(pool.apply_async(submit_job, (command+arguments,)))

    # Close the pool and wait for each running task to complete
    pool.close()
    pool.join()
    log_folder = None
    if (not options.output_folder is None):
        log_folder = str(options.output_folder)+"/"+"log"
    else:
        log_folder = "log"
    os.makedirs(log_folder, exist_ok=True)
    logfile = log_folder + "/" + str(options.run_name) + ".log"
    f = open(logfile, 'w')
    for result in results:
        out, err = result.get()
        errors = ["ERROR", "error"]
        print("out: {} err: {}".format(out, err), file=f)
        errmessage = "".join(map(chr, err))
        for error in errors:
            if(re.search(error, errmessage)):
                print("Something wrong occured. More info in the log file " + logfile)


if __name__ == '__main__':
    main()
