!----------------------------------------------------------------------------------------------------
! Program radish_example                                                                            !
!                                                                                                   !
! arXiv:1604.02191, arXiv:1705.09127, arXiv:1909.04704                                              !
! single and double-differential resummation for transverse observables up to N3LL accuracy         !
!                                                                                                   !
! authors: Wojciech Bizon, Pier Francesco Monni, Emanuele Re, Luca Rottoli, Paolo Torrielli         !
! developers: Pier Francesco Monni, Emanuele Re, Luca Rottoli, Paolo Torrielli                      !
!----------------------------------------------------------------------------------------------------
program radish_example
   use sub_defs_io; use types; use consts_dp
   use rad_tools; use pdfs_tools
   use expansion; use resummation; use reader
   use masters
   use opts_cmdline_cardfile
   use ew_parameters
   use mass_corr
   use matching
   use banner
   use warnings_and_errors
   use interpolation_radish
   use histogram_routines
   use seeds
   implicit none

   ! --------------------------------------------------------------------
   type(reader_opts)             :: opts, dummy_opts
   real(dp), allocatable, target :: v_and_sigmabar(:, :), resummed(:), resummed_error(:), &
                                 & resummed_expanded(:, :), matched(:), interp_nnlo(:), v_and_sigmabar_ref(:, :)
   real(dp), allocatable         :: dlumi_lumi(:), dlumi_lumi_expanded(:, :)
   real(dp), allocatable         :: final_result(:, :), diff_result(:, :), splb(:), splc(:), spld(:)
   real(dp)                      :: jet_radius, jet_veto, vmin, vmax, M, Q, muR, muF, p, rts, xQ
   real(dp)                      :: sigma(0:3), norm_resum, lumi_asympt(0:3)
   real(dp), pointer             :: v(:)
   integer                       :: pdf_member, ibin, nbins, order, iunit = 7, nev, ncall1, ncall2, iseq, i
   integer                       :: fixed_order
   real(dp)                      :: dy_error, ptmax_over_M, hmult
   character(len=100)            :: pdf_name
   character(len=2)              :: proc
   character(len=4)              :: hdecay_mode
   character(len=1)              :: flavour1, flavour2
   character(len=5)              :: collider
   character(len=20)             :: algo
   character(len=10)             :: matching_scheme
   character(len=200)            :: outfile, infile, interp_infile
   character(len=7)              :: loop_mass
   character(len=10)             :: observable
   logical                       :: dynamical, dryrun, prime
   logical                       :: differential, logscale, fivecol, use_cuts
   logical                       :: resum_only, do_matching, cross_section, analytic_expansion
   logical                       :: print_expansion_coefficients, print_resum_error
   logical                       :: fixed_flavour_scheme
   logical                       :: give_error = .true.
   logical                       :: radish_inclusive
   logical                       :: use_hoppet_running
   real(dp)                      :: ybottom, ycharm, ystrange
   real(dp)                      :: mmin, mmax, etamin, etamax
   integer, parameter            :: histo_index = 1
   logical                       :: singlepoint
   real(dp)                      :: M_fixed, Y_fixed
   real(dp)                      :: alphasPDF
   ! --------------------------------------------------------------------

   !------------------------Banners and help-------------------------------
   write (6, *)
   call print_banner
   write (6, *)
   call sleep(1)
   write (0, *) ''
   write (0, *) ' '//achar(27)//'[1;34mUse -h or -help to print help menu'//achar(27)//'[0m'
   write (0, *) ''
   write (0, *) ''
   if (log_val_opt("-h") .or. log_val_opt("-help")) then
      call print_help
      stop
   end if
   !----------------------------------------------------------------------


   !-------------------------Global options-------------------------------

   ! When present, this option allows to compute just the resummation (no matching with fixed-order is performed)
   resum_only = log_val_opt("-resum-only")

   ! Add the following option to get the resummed cross section instead of the efficiency.
   cross_section = log_val_opt("-cross-section", .false.)

   ! If use_cuts .true. [default], phase space cuts are applied
   ! The input variable -nocuts set use_cuts to false
   use_cuts = .not. (log_val_opt("-nocuts", .false.))

   ! When present, allows one to print differential distribution
   differential = log_val_opt('-differential', .false.)

   ! If true, inclusive calculation is used [default: .false.]
   radish_inclusive = log_val_opt('-radish_inclusive', .false.)

   ! If true, no shower - only the born-event generator is called [default: .false.]
   dryrun = log_val_opt('-dryrun', .false.)

   ! Set random generator seed
   iseq = int_val_opt("-iseed", 1189)
   call setseeds(iseq, .false.)

   ! Observable: default is pt_boson ptB; other observables available are pt_jet (ptj),
   ! phistar (phi), ptB with a veto on ptj (JV_ptB), ptj with a veto on ptB (JV_ptj)
   observable = string_val_opt("-observable", "ptB")

   ! The jet radius is that for any member of the longitudinally
   ! invariant generalised kt family (in so-called "inclusive" mode)
   ! with E-scheme recombination.
   jet_radius = dble_val_opts_or_cmdline(opts, 'R', 0.4_dp, give_error=give_error)

   ! The jet veto requirement (used for observable=JV_ptB and JV_ptj)
   jet_veto = dble_val_opts_or_cmdline(opts, 'JV', 30._dp, give_error=give_error)

   !----------------------------------------------------------------------


   !------------------Collider, process, PDF set and mass scale------------------

   ! Collider: 'pp' or 'ppbar'
   collider = trim(string_val_opts_or_cmdline(opts, "collider", "pp", give_error=give_error))

   ! Set the C.M. energy
   rts = dble_val_opts_or_cmdline(opts, "sqrts", 13000._dp, give_error=give_error)

   ! Process: Higgs production in gluon fusion ('H'), Neutral Drell-Yan production ('DY'),
   ! Charged Drell-Yan production ('Wp' or 'Wm')
   proc = trim(string_val_opts_or_cmdline(opts, "proc", "H", give_error=give_error))

   ! initialise default EW parameter
   call init_default()

   if (proc == "H") then
      M = dble_val_opts_or_cmdline(opts, "M", 125._dp, give_error=give_error)
   else if (proc == "DY") then
      M = MZ
   else if ((proc == 'Wp') .or. (proc == 'Wm')) then
      M = MW
   end if
   ! set the global mass of the colour singlet
   cs%M = M

   pdf_name = trim(string_val_opts_or_cmdline(opts, "pdf_name", &
      &                                     "NNPDF31_nnlo_as_0118", give_error=give_error))
   pdf_member = int_val_opts_or_cmdline(opts, "pdf_member", 0, give_error=give_error)

   ! Heavy-quark mass effects in Higgs production's virtual corrections (heavy top limit by default)
   loop_mass = trim(string_val_opts_or_cmdline(opts, "loop-mass", "none", give_error=give_error)) ! 't', 't+b' or 'none' (large-mt limit)

   ! Read bottom and charm yukawa modification (i.e. kappa) values
   ybottom = dble_val_opt('-yb', 1._dp)
   ycharm = dble_val_opt('-yc', 1._dp)
   ystrange = dble_val_opt('-ys', 1._dp)

   ! Initialise the Yukawa couplings
   call init_yukawas(ybottom, ycharm, ystrange)

   ! Higgs decay mode
   hdecay_mode = trim(string_val_opts_or_cmdline(opts, "hdecay", "none", give_error=give_error))

   !----------------------------------------------------------------------

   !---------------------------Set binning----------------------------------

   !--- Read fixed order file if option do_matching is set to true
   do_matching = .not. resum_only
   if (do_matching) then
      infile = string_val_opt("-in", "")
      if (infile == '') then
         write (0, *) ''//achar(27)//'[1;31mPlease specify fixed order file name with -in commandline option'//achar(27)//'[0m'
         stop
      endif
      vmax = dble_val_opt("-vmax", 1e200_dp)
      call read_file_into_array(infile, v_and_sigmabar, opts, max_col1=vmax)
      nbins = size(v_and_sigmabar, 2)

      ! Filter for possible infinities or NaNs
      do i = 1, nbins
         if (v_and_sigmabar(1, i) == zero) then
            ! If the v is set to zero in the fixed order file a NaN is produced by the code. This is a mistake
            ! in the fixed-order file that actually must have a IR cutoff
            call wae_error(''//achar(27)//'[1;31mRadish: found zero in v values of the fixed-order file'//achar(27)//'[0m')
         end if
      end do

      ! Total cross sections in nb
      sigma(0) = dble_val(opts, "xsct_lo")
      sigma(1) = dble_val(opts, "delta_xsct_nlo")
      sigma(2) = dble_val(opts, "delta_xsct_nnlo")
      sigma(3) = dble_val(opts, "delta_xsct_nnnlo", default=zero)

      ! Internal use only: interpolate the NNLO column using some auxiliary file
      interp_infile = string_val_opt("-interp", "-1")
      if (interp_infile .ne. "-1") then
         call read_file_into_array(interp_infile, v_and_sigmabar_ref, dummy_opts, max_col1=vmax)
         ! total cross sections in nb
         allocate (interp_nnlo(nbins), splb(nbins), splc(nbins), spld(nbins))
         write (*, *) "interpolating NNLO from reference file ..."
         do ibin = 1, nbins
            ! Interpolate using cubic spline
            ! call spline(v_and_sigmabar_ref(1,:), v_and_sigmabar_ref(3,:),splb,splc,spld,nbins)
            ! interp_nnlo(ibin) = ispline(v_and_sigmabar(1,ibin),v_and_sigmabar_ref(1,:), &
            !                & v_and_sigmabar_ref(3,:),splb,splc,spld,nbins)
            ! Interpolate using CERNLIB's polint
            call interpolate(v_and_sigmabar_ref(1, :), v_and_sigmabar_ref(3, :), &
                     & v_and_sigmabar(1, ibin), interp_nnlo(ibin), dy_error, 10)
         end do
      end if
   else
      ! Set binning
      vmin = dble_val_opt('-vmin', 1._dp)
      vmax = dble_val_opt('-vmax', 300._dp)
      nbins = int_val_opt('-nbins', 300);
      ! Create v_and_sigmabar with zero entries for the sigmabar
      allocate (v_and_sigmabar(3, nbins))
      v => v_and_sigmabar(1, :)
      if (log_val_opt('-log')) then
         do ibin = 1, nbins
            v(ibin) = exp(log(vmin) + (ibin - 1)*(log(vmax) - log(vmin))/(nbins - 1))
         end do
      else if (log_val_opt('-log10')) then
         do ibin = 1, nbins
            v(ibin) = 10**(log10(vmin) + (ibin - 1)*(log10(vmax) - log10(vmin))/(nbins - 1))
         end do
      else
         do ibin = 1, nbins
            v(ibin) = vmin + (ibin - 1)*(vmax - vmin)/(nbins - 1)
         end do
      end if
      v_and_sigmabar(2:, :) = zero
      sigma = zero
   end if

   !----------------------------------------------------------------------


   !------------------------------Set cuts--------------------------------

   mmin = dble_val_opts_or_cmdline(opts, "mmin", 10d0, give_error=give_error)
   mmax = dble_val_opts_or_cmdline(opts, "mmax", rts, give_error=give_error)
   etamin = dble_val_opts_or_cmdline(opts, "etamin", -1e99_dp, give_error=give_error)
   etamax = dble_val_opts_or_cmdline(opts, "etamax", 1e99_dp, give_error=give_error)

   !--------------------------------------------------------s--------------


   !----------------------------Scale options-------------------------------

   !--- Check if using dynamical scale [default = .false. ]
   dynamical = log_val_opt("-dynamical", .false.)

   ! Renormalisation, factorisation and resummation scales
   Q = var_or_xVar("Q", xdefault=0.5_dp, opts=opts, M=M, override_allowed=.true.)
   muR = var_or_xVar("muR", xdefault=1.0_dp, opts=opts, M=M, override_allowed=.false.) ! must be consistent with fixed-order
   muF = var_or_xVar("muF", xdefault=1.0_dp, opts=opts, M=M, override_allowed=.false.) ! must be consistent with fixed-order

   !----------------------------------------------------------------------


   !-------------Resummation and matching expansion options-----------------------

   ! The resummation order (0 = LL, 1 = NLL, 2 = NNLL [default], 3 = N3LL)
   order = int_val_opt("-order", order_NNLL)

   ! If present, uses prime counting
   prime = log_val_opt('-prime', .false.)

   ! The fixed order (1=NLO, 2=NNLO [default], 3=N3LO)
   fixed_order = int_val_opt('-fixed_order', fixed_order_NNLO)

   ! The power for the damping of the logarithm at large v (see Eq.4.3 of 1705.09127)
   p = dble_val_opts_or_cmdline(opts, "p", 5._dp);

   ! The matching scheme
   matching_scheme = string_val_opt("-scheme", "multa")

   !----------------------------------------------------------------------


   !--------- Initialise PDFs and evolution [default: LHAPDF] -----------

   ! If present, used a fixed five flavour scheme
   fixed_flavour_scheme = log_val_opt("-fixed-flavour-scheme")

   if (fixed_flavour_scheme) then
      call init_pdfs_custom_evolution(trim(pdf_name), pdf_member, rts, order)
   else
      call init_pdfs_from_LHAPDF(trim(pdf_name), pdf_member)
   end if

   ! If present, used the hoppet running rather than the LHAPDF one
   use_hoppet_running = log_val_opt("-use_hoppet_running", .false.)

   !----------------------------------------------------------------------


   !----------- Decide where to send output and in what format --------------

   outfile = string_val_opt("-out", "/dev/stdout")
   if (outfile == "-") then
      iunit = 6
   else
      open (7, file=outfile)
   end if

   !----------------------------------------------------------------------


   !--------------------------Shower and vegas events-----------------------------

   ! Number of shower events for the numerical calculation
   if (radish_inclusive) then
      nev = nint(dble_val_opt('-nev', 1e7_dp))
   else
      nev = nint(dble_val_opt('-nev', 1e3_dp))
   endif

   ! Number of Born events to be generated (used if radish_inclusive is false)
   ncall1 = nint(dble_val_opt('-ncall1', 1e3_dp)) ! warm up
   ncall2 = nint(dble_val_opt('-ncall2', 1e3_dp)) ! main run

   !----------------------------------------------------------------------


   !------------------------Extra (debug) options---------------------------

   ! When present, this option allows one to print the individual expansion coefficients in the output file
   print_expansion_coefficients = log_val_opt("-print-expansion-coefficients")

   ! When present, this option allows one to print the resummed error
   print_resum_error = log_val_opt("-print-resum-error")

   !--- decide whether to use analytic expansion WITHOUT modified logarithms
   !--- and valid only for central scales (this is meant for checks only).
   !--- This routine has been implemented only at N3LL
   analytic_expansion = log_val_opt("-analytic-expansion")

   !--- Decide what initial-state flavour to consider in the expansion (ONLY)
   !--- parton1 (from h1), parton2 (from h2) -> qq, qx, xq, qg, gq, xg, gx, gg, xx
   !--- To run this feature, set flavour_breakdown=.true. rad_tools.f90 (Testing only)
   flavour1 = string_val_opt("-flavour1", "q")
   flavour2 = string_val_opt("-flavour2", "q")

   ! if true, use fixed Y and M
   ! WARNING: Wrong normalisation of the curve
   singlepoint = log_val_opt('-singleYMpoint', .false.)
   M_fixed = dble_val_opt('-Mfixed', 91.187d0)
   Y_fixed = dble_val_opt('-Yfixed', 0d0)

   ! Multiplicative matching parameters - modify them if you know what you are doing (note: not affecting multa scheme)
   ptmax_over_M = dble_val_opt("-ptmax_over_M", 0.5_dp)
   hmult = dble_val_opt("-hmult", three)

   ! The following call is relevant for certain internal tests.
   ! Normal users should not need it.
   ! call small_alpha_init()

   !----------------------------------------------------------------------


   !----------------------------------------------------------------------

   ! Verify we've used all the command-line arguments and all the options
   ! from the fixed-order file
   if (.not. CheckAllArgsUsed(0)) stop
   call assert_all_opts_used(opts)

   ! Now initialise "cs" which is the object that contains the coefficients
   ! and is used also to communicate information on the collider,
   ! scales, etc., throughout the rest of the program
   call set_process_and_parameters(collider, proc, rts, M, muR, muF, Q, p, order, mmin, mmax, etamin, etamax, &
         &jet_radius, jet_veto, loop_mass, observable, dynamical, dryrun, analytic_expansion, flavour1, flavour2, hdecay_mode, use_hoppet_running)

   ! This call arranges for the A and B resummation coefficients to be set
   call init_proc()

   if (prime) then
      call set_prime_counting()
   endif

   if (singlepoint) then
      call set_fixed_M_and_Y(M_fixed, Y_fixed)
   endif

   ! internal use only: overwrite the NNLO column with the interpolated one
   if ((loop_mass .ne. "none") .and. (interp_infile .ne. "-1")) then
      v_and_sigmabar(3, :) = interp_nnlo
   end if

   !-- perform the resummation, matching, and output of the results
   if (print_expansion_coefficients) then
      if (print_resum_error) then
         allocate (final_result(9, 0:nbins))
         if (differential) allocate (diff_result(9, 0:nbins - 1))
      else
         allocate (final_result(8, 0:nbins))
         if (differential) allocate (diff_result(8, 0:nbins - 1))
      endif
   else
      if (print_resum_error) then
         allocate (final_result(6, 0:nbins))
         if (differential) allocate (diff_result(6, 0:nbins - 1))
      else
         allocate (final_result(5, 0:nbins))
         if (differential) allocate (diff_result(5, 0:nbins - 1))
      endif
   end if

   allocate (resummed(nbins), resummed_error(nbins), resummed_expanded(0:3, nbins), matched(nbins),&
         &dlumi_lumi(nbins), dlumi_lumi_expanded(2, nbins))

   ! Initialise expansion:
   ! The Master Integrals are defined for the observable itself
   call init_expansion()

   ! Allocate histograms (e.g. 1 in the following example)
   call book_histogram(cs%observable, cs%observable, histo_index, v_and_sigmabar(1, :))
   ! Set the index that represents the scale (e.g. 1 for a single scale per run)
   call set_scale_index(histo_index)
   call set_binvar_index(histo_index)
   call allocate_master_integrals(cs%observable, cs%observable, size(histo_resummed(histo_index)%values%right(:)))

   ! The following calls get the resummation and its expansion, and
   ! also set dlumi_lumi and dlumi_lumi_expanded, which are
   ! ingredients for the matching below
   if (radish_inclusive) then
      call resummed_sigma_inclusive(order, nev)
   else
      call resummed_sigma(order, nev, ncall1, ncall2, iseq, use_cuts)
   end if

   resummed(:) = histo_resummed(histo_index)%values%weight(:)
   resummed_error(:) = histo_resummed(histo_index)%values%error(:)
   resummed_expanded(0, :) = histo_expanded(histo_index)%values%lo(:)
   resummed_expanded(1, :) = histo_expanded(histo_index)%values%nlo(:)
   resummed_expanded(2, :) = histo_expanded(histo_index)%values%nnlo(:)
   resummed_expanded(3, :) = histo_expanded(histo_index)%values%n3lo(:)
   lumi_asympt(0) = histo_luminosity(histo_index)%values%asympt
   lumi_asympt(1) = histo_luminosity(histo_index)%values%as
   lumi_asympt(2) = histo_luminosity(histo_index)%values%as2
   lumi_asympt(3) = histo_luminosity(histo_index)%values%as3

   norm_resum = resummed_expanded(0, 1) ! This sets the cross section units
   if (do_matching) then
      resummed = resummed*sigma(0)/norm_resum
      resummed_error = resummed_error*sigma(0)/norm_resum
      if (.not. flavour_breakdown) resummed_expanded = resummed_expanded*sigma(0)/norm_resum
      lumi_asympt = lumi_asympt*sigma(0)/norm_resum
      matched = match(matching_scheme,&
            &  ptmax_over_M, hmult, &
            &  sigma, v_and_sigmabar(1, :), resummed, resummed_expanded(:fixed_order, :), &
            &  v_and_sigmabar(2:fixed_order + 1, :), dlumi_lumi, dlumi_lumi_expanded, lumi_asympt)

      ! -- Normalize efficiency for schemes a and b
      if (matching_scheme == 'a') then
         matched = matched/sum(sigma(0:fixed_order))
      elseif (matching_scheme == 'b') then
         matched = matched/sum(sigma(0:fixed_order - 1))
      elseif (matching_scheme == 'c') then
         matched = matched/sum(sigma(0:fixed_order - 2))
      elseif (matching_scheme == 'd') then
         matched = matched/sum(sigma(0:fixed_order - 3))
      elseif (matching_scheme == 'R') then
         matched = matched/sum(sigma(0:fixed_order))
      elseif (matching_scheme == 'mult') then
         matched = matched/sum(sigma(0:fixed_order))
      elseif (matching_scheme == 'multa') then
         matched = matched/sum(sigma(0:fixed_order))
      endif
   else
      sigma(0) = resummed_expanded(0, 1)
      sigma(1:) = zero
      matched = zero
   end if

   !----------------------------------------------------------------------

   ! Send documentation to the output files
   write (iunit, '(a)') "# "//trim(command_line())
   call print_parameters(iunit)
   write (iunit, '(a)') "# order = "//order_string(order)
   write (iunit, '(a,i4)') "# pdf name and set = "//trim(pdf_name)//", ", pdf_member
   write (iunit, '(a)') "# matching scheme = "//trim(matching_scheme)
   write (iunit, '(a,i15)') '# observable= '//trim(cs%observable)
   write (iunit, '(a,i15)') '# nev=', nev
   if (do_matching) then
      write (iunit, '(a,4f16.6)') '# total cross sections at LO, NLOdiff, NNLOdiff, N3LOdiff = ', sigma
   else
      write (iunit, '(a,8es18.8e3)') '# native cross sections at LO = ', sigma(0)
   endif

   ! v
   final_result(1, 1:) = v_and_sigmabar(1, :)
   ! matched
   final_result(2, 1:) = matched
   ! pure resummed
   final_result(3, 1:) = resummed(:)/(sigma(0) + sum(sigma(1:order - 1))) ! arbirtary normalization

   ! pure fixed order (in a scheme analogous to that used for the matching).
   final_result(4, 1:) = fixed_order_schemes(matching_scheme, sigma, v_and_sigmabar(2:fixed_order + 1, :))

   ! Print the expansion (for internal checks)
   final_result(5, 1:) = (sigma(0) + resummed_expanded(1, :))/(sigma(0) + sum(sigma(1:order - 1)))
   if (fixed_order > 1) then
      final_result(5, 1:) = (sigma(0) + resummed_expanded(1, :) + resummed_expanded(2, :))&
            & /(sigma(0) + sum(sigma(1:order - 1)))
   end if
   if (fixed_order > 2) then
      final_result(5, 1:) = (sigma(0) + resummed_expanded(1, :) + resummed_expanded(2, :) + resummed_expanded(3, :))&
            & /(sigma(0) + sum(sigma(1:order - 1)))
   end if

   if (print_expansion_coefficients) then
      final_result(6, 1:) = (resummed_expanded(1, :))
      final_result(7, 1:) = (resummed_expanded(2, :))
      final_result(8, 1:) = (resummed_expanded(3, :))
      if (print_resum_error) then
         final_result(9, 1:) = resummed_error(:)/(sigma(0) + sum(sigma(1:order - 1)))
      endif
   else
      if (print_resum_error) then
         final_result(6, 1:) = resummed_error(:)/(sigma(0) + sum(sigma(1:order - 1)))
      endif
   end if

   ! Value at pt=zero
   final_result(1:, 0) = zero ! WARNING: This assumption is wrong for expansion and fixed-order columns

   if (flavour_breakdown) then
      ! Print out the expansion coefficients separately for the channel breakdown
      write (iunit, *) '# Printing the ', flavour1, flavour2, ' channel at fixed order', fixed_order
      if (fixed_order == 1) final_result(5, 1:) = resummed_expanded(1, :)
      if (fixed_order == 2) final_result(5, 1:) = resummed_expanded(2, :)
      if (fixed_order == 3) final_result(5, 1:) = resummed_expanded(3, :)
   end if

   if (cross_section) then
      if (matching_scheme == 'a' .or. matching_scheme == 'mult' .or. matching_scheme == 'multa' .or. matching_scheme == 'R') then
         final_result(2, :) = final_result(2, :)*sum(sigma(0:fixed_order))
         final_result(3, :) = final_result(3, :)*(sigma(0) + sum(sigma(1:order - 1)))
         final_result(4, :) = final_result(4, :)*sum(sigma(0:fixed_order))
         final_result(5, :) = final_result(5, :)*(sigma(0) + sum(sigma(1:order - 1)))
         if (print_resum_error) then
            if (print_expansion_coefficients) then
               final_result(9, :) = final_result(9, :)*(sigma(0) + sum(sigma(1:order - 1)))
            else
               final_result(6, :) = final_result(6, :)*(sigma(0) + sum(sigma(1:order - 1)))
            endif
         endif
      else
         call wae_error(''//achar(27)//'[1;31mradish: cross_section option not compatible with the selected matching scheme'//achar(27)//'[0m')
      end if
   endif

   ! Add lines to allow for matching
   if (print_expansion_coefficients) then
      write (iunit, '(a,8es18.8e3)') "# exp0=", resummed_expanded(0, 1)
      write (iunit, '(a,8es18.8e3)') "# lumi_asympt0=", lumi_asympt(0)
      write (iunit, '(a,8es18.8e3)') "# lumi_asympt1=", lumi_asympt(1)
      write (iunit, '(a,8es18.8e3)') "# lumi_asympt2=", lumi_asympt(2)
   end if
   ! Put out a header for the numbers
   write (iunit, '(a)') '#----------------------------------------------------------------------'
   if (print_expansion_coefficients) then
      if (print_resum_error) then
         write(iunit,'(a)') '#         v            matched      resummed     fixed-order    expansion             exp1                 exp2                exp3           resummed_error'
      else
         write(iunit,'(a)') '#         v            matched      resummed     fixed-order    expansion             exp1                 exp2                exp3'
      endif
   else
      if (print_resum_error) then
         write (iunit, '(a)') '#         v            matched      resummed     fixed-order    expansion         resummed_error'
      else
         write (iunit, '(a)') '#         v            matched      resummed     fixed-order    expansion'
      endif
   end if

   !---- Final output of the results
   if (differential) then
      ! old binning
      !final_result(2:,1:nbins-1) = (final_result(2:,2:nbins)-final_result(2:,1:nbins-1))/&
      !     &             spread(final_result(1,2:nbins)-final_result(1,1:nbins-1),ncopies=size(final_result,dim=1)-1,dim=1)
      !final_result(1,1:nbins-1) = (final_result(1,2:nbins)+final_result(1,1:nbins-1))/2
      ! PM new binning
      ! final_result(2:,1:nbins-1) = (final_result(2:,1:nbins)-final_result(2:,0:nbins-1))/&
      !      &             spread(final_result(1,1:nbins)-final_result(1,0:nbins-1),ncopies=size(final_result,dim=1)-1,dim=1)
      ! final_result(1,1:nbins-1) = (final_result(1,1:nbins)+final_result(1,0:nbins-1))/2
      do ibin = 1, nbins
         diff_result(2:, ibin - 1) = (final_result(2:, ibin) - final_result(2:, ibin - 1))/&
               &             spread(final_result(1, ibin) - final_result(1, ibin - 1), ncopies=size(final_result, dim=1) - 1, dim=1)
         diff_result(1, ibin - 1) = (final_result(1, ibin) + final_result(1, ibin - 1))/2
         ! Set fixed-order and expansion to 0 in the very first bin to avoid confusion
         if (print_resum_error) then
            if (print_expansion_coefficients) then
               diff_result(4:8, 0) = zero
            else
               diff_result(4:5, 0) = zero
            endif
         else
            diff_result(4:, 0) = zero
         endif
         write (iunit, '(9es18.8e3)') diff_result(:, ibin - 1)
      end do
   else
      do ibin = 1, nbins
         write (iunit, '(9es18.8e3)') final_result(:, ibin)
      end do
   end if
   if (iunit /= 6) close (iunit)

   if (outfile /= "-") then
      write (0, *) '                                                                      '
      write (0, *) ''//achar(27)//'[1;94mResults saved in file: ', outfile//achar(27)//'[0m'
   end if

   ! Deallocate memory
   if (do_matching) then
      if (interp_infile .ne. "-1") then
         deallocate (interp_nnlo, splb, splc, spld)
      end if
   else
      deallocate (v_and_sigmabar)
   end if

   deallocate (final_result)
   if (differential) deallocate (diff_result)

   deallocate (resummed, resummed_error, resummed_expanded, matched,&
         &dlumi_lumi, dlumi_lumi_expanded)

   call terminate_expansion()
   call deallocate_histograms()
   call deallocate_grids()

   ! Coefficient functions are deallocated only if order >= NLL'
   if (order > order_NLL .or. (order == order_NLL .and. cs%prime)) then
      call deallocate_C_matxs()
   endif

end program radish_example
