module matching
  use types; use consts_dp
  use warnings_and_errors
  use pdfs_tools; use rad_tools

  implicit none

  private
  public :: match, fixed_order_schemes
  public :: matching_mult_a, matching_R  !TODO: this is needed for the matrix interface - decide if this is okay
  public :: matching_mult_a_matrix, matching_R_matrix  !TODO: this is needed for the matrix interface - decide if this is okay
  public :: matching_mult_damping  !TODO: this is needed for the matrix interface - decide if this is okay
  integer, public, parameter :: fixed_order_NLO = 1, fixed_order_NNLO = 2, fixed_order_N3LO = 3

contains

  !----------------------------------------------------------------------
  function match(matching_scheme, &
       &            ptmax_over_M, hmult, &
       &            sigma, pt, resummation, resum_expanded, &
       &            sigmabar,dlumi_lumi,dlumi_lumi_expanded, lumi_asympt) result(res)
    character(len=*), intent(in) :: matching_scheme
    real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:), lumi_asympt(0:3)
    real(dp), intent(in) :: dlumi_lumi(:), dlumi_lumi_expanded(:,:)
    real(dp), intent(in) :: ptmax_over_M, hmult
    real(dp)             :: res(size(pt))

    select case(trim(matching_scheme))
    case("a")
       res = matching_a(sigma, pt(:), resummation, resum_expanded, &
            &sigmabar(:,:),dlumi_lumi,dlumi_lumi_expanded)
    case("b")
       res = matching_b(sigma, pt(:), resummation, resum_expanded, &
            &sigmabar(:,:),dlumi_lumi,dlumi_lumi_expanded)
    case("c")
       res = matching_c(sigma, pt(:), resummation, resum_expanded, &
            &sigmabar(:,:),dlumi_lumi,dlumi_lumi_expanded)
    case("d")
       res = matching_d(sigma, pt(:), resummation, resum_expanded, &
            &sigmabar(:,:),dlumi_lumi,dlumi_lumi_expanded)
    case("R")
       res = matching_R(sigma, pt(:), resummation, resum_expanded, sigmabar)
    case("mult")
       res = matching_mult(sigma, pt(:), resummation, resum_expanded, &
            &sigmabar(:,:), lumi_asympt, ptmax_over_M, hmult)
    case("multa")
       res = matching_mult_a(sigma, pt(:), resummation, resum_expanded, &
            &sigmabar(:,:), lumi_asympt, ptmax_over_M, hmult)
    case default
       call wae_error("matching: unrecognised matching scheme "//matching_scheme)
    end select
  end function match



  ! --- new version (designed for ptB numerical resummation)
  ! We compute (Sigma_RES/sigma(0))^Z*(sigma(0)^Z*Sigma_FO/Sigma_EXP^Z + O(as^(n+1)))
  ! where n is the order we're matching onto.
  ! In the pt->inf limit this tends to (Z=1 for simplicity)
  ! normalization=Lumi_RES(pt->inf)/sigma(0)*(sigma(0)+sigma(1) - Lumi(1) + ...)
  ! which differs from the total inclusive XS by higher-order terms, i.e. O(as^(n+1)).
  ! We then restore the correct normalisation by multiplying by
  ! sigma_tot/normalization = 1 + O(as^(n+1)).
  ! This rescaling introduces a correction to the logarithmic tower that is formally N^(n+2)LL:
  ! i.e. for NLO matching is O(as^2) (this can be critical if we want to control terms O(as^3*L^2))
  ! for NNLO matching is O(as^3), for N3LO matching is O(as^4). So for the latter two cases we're not
  ! spoiling the NNLL accuracy by any means.
  ! We also introduce an additional factor Z = (1-pt/ptmax) that kills the
  ! resummation at large pt.
  ! Z=1 recovers the standard multiplicative matching.
  ! In the asymptotic limit, it normalizes to (without damping factor Z):
  ! L(muF)/sigma(0)*[sigma(0) + sigma(1) - L10(muF) +  !! O(as)
  ! + sigma(2) - L20(muF) + L10(muF)^2/sigma(0) - L10(muF)*sigma(1)/sigma(0) + !! O(as^2)
  ! + sigma(3) - L30(muF) - L10^3(muF)/sigma(0)^2 + L10(muF)^2*sigma(1)/sigma(0)^2 - L20(muF)*sigma(1)/sigma(0) + !! O(as^3)
  ! + L10(muF)*(2*L20(muF) - sigma(2))/sigma(0)]
  ! Notice that when a damping factor Z is introduced, the normalisation is simply
  ! sigma(0) + sigma(1) + sigma(2) + sigma(3), since in the code below zeta=Z(infinity)=0
  function matching_mult(sigma, pt, resummation, resum_expanded, sigmabar, lumi_asympt, ptmax_over_M, hmult) result(res)
    real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:), lumi_asympt(0:3)
    real(dp), intent(in) :: hmult, ptmax_over_M
    real(dp)             :: res(size(pt))
    !----------------------------------------------
    integer              :: order
    real(dp)             :: matching_factor(size(pt)), Z(size(pt)), ptmax, normalization, zeta

    order = size(resum_expanded,dim=1) - 1 ! resum_expanded has fixed_order+1 dimensions

    ! get the damping factor
    ptmax = cs%M*ptmax_over_M
    if (cs%observable=='phi') ptmax = ptmax/cs%M ! this ratio is kept constant in the generation
    Z = (one - pt/ptmax)**hmult ! choose the power > 2 to have a smooth second derivative
    where (pt >= ptmax) Z = zero

    ! no damping factor
    !Z=one
    zeta=Z(size(Z))

    matching_factor = sigma(0) + sigma(1)+sigmabar(1,:) -Z*resum_expanded(1,:)
    normalization = sigma(0) + sigma(1) - zeta*lumi_asympt(1)

    if (order >= 2) then
       matching_factor = matching_factor &
            &          + sigma(2) + sigmabar(2,:) - Z*resum_expanded(2,:) &
            &          + Z*(one+Z)/two*resum_expanded(1,:)**2/sigma(0) &
            &          - Z*resum_expanded(1,:)*(sigma(1)+sigmabar(1,:))/sigma(0)
       normalization = normalization + sigma(2) - zeta*lumi_asympt(2) + zeta*(one+zeta)/two*lumi_asympt(1)**2/sigma(0) &
            &          - zeta*lumi_asympt(1)*sigma(1)/sigma(0)
    end if

    if (order >= 3) then
       matching_factor = matching_factor &
            &          + sigma(3) + sigmabar(3,:) - Z*resum_expanded(3,:) &
            &          - Z*(one+Z)*(two+Z)/6._dp*resum_expanded(1,:)**3/sigma(0)**2 &
            &          + Z*(one+Z)/two*resum_expanded(1,:)**2*(sigma(1)+sigmabar(1,:))/sigma(0)**2 &
            &          - Z*resum_expanded(2,:)*(sigma(1)+sigmabar(1,:))/sigma(0) &
            &          + Z*resum_expanded(1,:)*((one+Z)*resum_expanded(2,:) &
            &          - sigma(2) - sigmabar(2,:))/sigma(0)
       normalization = normalization + sigma(3) - zeta*lumi_asympt(3) - zeta*(one+zeta)*(two+zeta)/6._dp*lumi_asympt(1)**3/sigma(0)**2 &
            & + zeta*(one+zeta)/two*lumi_asympt(1)**2*sigma(1)/sigma(0)**2 &
            & - zeta*lumi_asympt(2)*sigma(1)/sigma(0) &
            & + zeta*lumi_asympt(1)*((one+zeta)*lumi_asympt(2) - sigma(2))/sigma(0)
    end if
    res =  (resummation/sigma(0))**Z*matching_factor
    ! reset proper normalization
    normalization = (lumi_asympt(0)/sigma(0))**zeta * normalization
    res = res / normalization * sum(sigma(0:order))
  end function matching_mult


  function matching_mult_damping(sigma, pt, resummation, resum_expanded, sigmabar, lumi_asympt, ptmax_over_M, hmult, observable, scale) result(res)
    real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:), lumi_asympt(0:3)
    real(dp), intent(in) :: hmult, ptmax_over_M
    real(dp)             :: res(size(pt))
    !----------------------------------------------
    integer              :: order
    real(dp)             :: matching_factor(size(pt)), Z(size(pt)), ptmax, normalization, zeta
    character(len=*), intent(in) :: observable
    real(dp), intent(in), value:: scale

    order = size(resum_expanded,dim=1) - 1 ! resum_expanded has fixed_order+1 dimensions

    ! get the damping factor
    ptmax = scale*ptmax_over_M
    if (observable=='phi') ptmax = ptmax/scale ! this ratio is kept constant in the generation
    Z = (one - pt/ptmax)**hmult ! choose the power > 2 to have a smooth second derivative
    where (pt >= ptmax) Z = zero

    ! no damping factor
    !Z=one
    zeta=Z(size(Z))

    matching_factor = sigma(0) + sigma(1)+sigmabar(1,:) -Z*resum_expanded(1,:)
    normalization = sigma(0) + sigma(1) - zeta*lumi_asympt(1)

    if (order >= 2) then
       matching_factor = matching_factor &
            &          + sigma(2) + sigmabar(2,:) - Z*resum_expanded(2,:) &
            &          + Z*(one+Z)/two*resum_expanded(1,:)**2/sigma(0) &
            &          - Z*resum_expanded(1,:)*(sigma(1)+sigmabar(1,:))/sigma(0)
       normalization = normalization + sigma(2) - zeta*lumi_asympt(2) + zeta*(one+zeta)/two*lumi_asympt(1)**2/sigma(0) &
            &          - zeta*lumi_asympt(1)*sigma(1)/sigma(0)
    end if

    if (order >= 3) then
       matching_factor = matching_factor &
            &          + sigma(3) + sigmabar(3,:) - Z*resum_expanded(3,:) &
            &          - Z*(one+Z)*(two+Z)/6._dp*resum_expanded(1,:)**3/sigma(0)**2 &
            &          + Z*(one+Z)/two*resum_expanded(1,:)**2*(sigma(1)+sigmabar(1,:))/sigma(0)**2 &
            &          - Z*resum_expanded(2,:)*(sigma(1)+sigmabar(1,:))/sigma(0) &
            &          + Z*resum_expanded(1,:)*((one+Z)*resum_expanded(2,:) &
            &          - sigma(2) - sigmabar(2,:))/sigma(0)
       normalization = normalization + sigma(3) - zeta*lumi_asympt(3) - zeta*(one+zeta)*(two+zeta)/6._dp*lumi_asympt(1)**3/sigma(0)**2 &
            & + zeta*(one+zeta)/two*lumi_asympt(1)**2*sigma(1)/sigma(0)**2 &
            & - zeta*lumi_asympt(2)*sigma(1)/sigma(0) &
            & + zeta*lumi_asympt(1)*((one+zeta)*lumi_asympt(2) - sigma(2))/sigma(0)
    end if
    res =  (resummation/sigma(0))**Z*matching_factor
    ! reset proper normalization
    normalization = (lumi_asympt(0)/sigma(0))**zeta * normalization
    res = res / normalization * sum(sigma(0:order))
  end function matching_mult_damping

  ! The following scheme is a variant of the (a) scheme for Jet Veto efficiencies, expressed in terms of
  ! cross sections and dividing out the whole luminosity at muF.
  function matching_mult_a(sigma, pt, resummation, resum_expanded, sigmabar, lumi_asympt, ptmax_over_M, hmult) result(res)
    real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:), lumi_asympt(0:3)
    real(dp), intent(in) :: hmult, ptmax_over_M
    real(dp)             :: res(size(pt))
    !----------------------------------------------
    integer              :: order, i
    real(dp)             :: matching_factor(size(resummation)), Z(size(pt)), ptmax
    logical, parameter   :: damping=.false.

    order = size(resum_expanded,dim=1) - 1

    ! get the damping factor
    if (damping) then
       ptmax = cs%M*ptmax_over_M
       if (cs%observable=='phi') ptmax = ptmax/cs%M ! this ratio is kept constant in the generation
       Z = (one - pt/ptmax)**hmult ! choose the power > 2 to have a smooth second derivative
       where (pt >= ptmax) Z = zero
    else
       Z = one
    end if

    matching_factor = sigma(0)*(lumi_asympt(0)/sigma(0))**Z + sigma(1)+sigmabar(1,:) - Z*resum_expanded(1,:)

    if (order >= 2) then
       matching_factor = matching_factor &
            &          + sigma(2) + sigmabar(2,:) - Z*resum_expanded(2,:) &
            &          + Z*lumi_asympt(1)/sigma(0)*(sigma(1)+sigmabar(1,:)) &
            &          + Z*(one+Z)/two*resum_expanded(1,:)**2/sigma(0) &
            &          - Z*resum_expanded(1,:)/sigma(0)*(sigma(1)+sigmabar(1,:)+Z*lumi_asympt(1))
    end if

    if (order >= 3) then
       matching_factor = matching_factor &
            &          + sigma(3) + sigmabar(3,:) - Z*resum_expanded(3,:) &
            &          - Z*(one+Z)*(two+Z)/6._dp*resum_expanded(1,:)**3/sigma(0)**2 &
            &          + Z*(one+Z)/two*resum_expanded(1,:)**2/sigma(0)**2*(sigma(1)+sigmabar(1,:) + Z*lumi_asympt(1)) &
            &          + Z/two*(two*sigma(0)*(sigma(1)+sigmabar(1,:))*(lumi_asympt(2)-resum_expanded(2,:)) &
            &          + (Z-one)*lumi_asympt(1)**2*(sigma(1)+sigmabar(1,:)) + two*lumi_asympt(1)*sigma(0)*(sigma(2)+sigmabar(2,:) - Z*resum_expanded(2,:)))/sigma(0)**2 &
            &          - Z/two*resum_expanded(1,:)*(Z*lumi_asympt(1)*(two*(sigma(1)+sigmabar(1,:)) + (Z-one)*lumi_asympt(1)) &
            &          + two*sigma(0)*(sigma(2)+sigmabar(2,:) + Z*lumi_asympt(2) - (one+Z)*resum_expanded(2,:)))/sigma(0)**2
    end if

    res =  (resummation/lumi_asympt(0))**Z*matching_factor

  end function matching_mult_a

  ! The following scheme is a variant of the (a) scheme for Jet Veto efficiencies, expressed in terms of
  ! cross sections and dividing out the whole luminosity at muF.
  function matching_mult_a_damping(sigma, pt, resummation, resum_expanded, sigmabar, lumi_asympt, ptmax_over_M, hmult, observable, scale) result(res)
   real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:), lumi_asympt(0:3)
   real(dp), intent(in) :: hmult, ptmax_over_M
   real(dp)             :: res(size(pt))
   !----------------------------------------------
   integer              :: order, i
   real(dp)             :: matching_factor(size(resummation)), Z(size(pt)), ptmax
   logical, parameter   :: damping=.true.
   character(len=*), intent(in) :: observable
   real(dp), intent(in), value:: scale

   order = size(resum_expanded,dim=1) - 1

   ! get the damping factor
   if (damping) then
      ptmax = scale*ptmax_over_M
      if (observable=='phi') ptmax = ptmax/scale ! this ratio is kept constant in the generation
      Z = (one - pt/ptmax)**hmult ! choose the power > 2 to have a smooth second derivative
      where (pt >= ptmax) Z = zero
   else
      Z = one
   end if

   matching_factor = sigma(0)*(lumi_asympt(0)/sigma(0))**Z + sigma(1)+sigmabar(1,:) - Z*resum_expanded(1,:)

   if (order >= 2) then
      matching_factor = matching_factor &
           &          + sigma(2) + sigmabar(2,:) - Z*resum_expanded(2,:) &
           &          + Z*lumi_asympt(1)/sigma(0)*(sigma(1)+sigmabar(1,:)) &
           &          + Z*(one+Z)/two*resum_expanded(1,:)**2/sigma(0) &
           &          - Z*resum_expanded(1,:)/sigma(0)*(sigma(1)+sigmabar(1,:)+Z*lumi_asympt(1))
   end if

   if (order >= 3) then
      matching_factor = matching_factor &
           &          + sigma(3) + sigmabar(3,:) - Z*resum_expanded(3,:) &
           &          - Z*(one+Z)*(two+Z)/6._dp*resum_expanded(1,:)**3/sigma(0)**2 &
           &          + Z*(one+Z)/two*resum_expanded(1,:)**2/sigma(0)**2*(sigma(1)+sigmabar(1,:) + Z*lumi_asympt(1)) &
           &          + Z/two*(two*sigma(0)*(sigma(1)+sigmabar(1,:))*(lumi_asympt(2)-resum_expanded(2,:)) &
           &          + (Z-one)*lumi_asympt(1)**2*(sigma(1)+sigmabar(1,:)) + two*lumi_asympt(1)*sigma(0)*(sigma(2)+sigmabar(2,:) - Z*resum_expanded(2,:)))/sigma(0)**2 &
           &          - Z/two*resum_expanded(1,:)*(Z*lumi_asympt(1)*(two*(sigma(1)+sigmabar(1,:)) + (Z-one)*lumi_asympt(1)) &
           &          + two*sigma(0)*(sigma(2)+sigmabar(2,:) + Z*lumi_asympt(2) - (one+Z)*resum_expanded(2,:)))/sigma(0)**2
   end if

   res =  (resummation/lumi_asympt(0))**Z*matching_factor

 end function matching_mult_a_damping

  ! The following scheme is a variant of the (a) scheme for Jet Veto efficiencies, expressed in terms of
  ! cross sections and dividing out the whole luminosity at muF.
  function matching_mult_a_matrix(sigmatot, pt, resummation, resum_expanded, sigma, lumi_asympt, ptmax_over_M, hmult) result(res)
   real(dp), intent(in) :: sigmatot, pt(:), resummation(:), resum_expanded(0:,:), sigma(1:,:), lumi_asympt(0:3)
   real(dp), intent(in) :: hmult, ptmax_over_M
   real(dp)             :: res(size(pt))
   !----------------------------------------------
   integer              :: order, i
   real(dp)             :: matching_factor(size(resummation)), Z(size(pt)), ptmax
   logical, parameter   :: damping=.false.

   order = size(resum_expanded,dim=1) - 1

   ! get the damping factor
   if (damping) then
      ptmax = cs%M*ptmax_over_M
      if (cs%observable=='phi') ptmax = ptmax/cs%M ! this ratio is kept constant in the generation
      Z = (one - pt/ptmax)**hmult ! choose the power > 2 to have a smooth second derivative
      where (pt >= ptmax) Z = zero
   else
      Z = one
   end if

   matching_factor = sigmatot*(lumi_asympt(0)/sigmatot)**Z + sigma(1,:) - Z*resum_expanded(1,:)

   if (order >= 2) then
      matching_factor = matching_factor &
           &          + sigma(2,:) - Z*resum_expanded(2,:) &
           &          + Z*lumi_asympt(1)/sigmatot*(sigma(1,:)) &
           &          + Z*(one+Z)/two*resum_expanded(1,:)**2/sigmatot &
           &          - Z*resum_expanded(1,:)/sigmatot*(sigma(1,:)+Z*lumi_asympt(1))
   end if

   if (order >= 3) then
      matching_factor = matching_factor &
           &          + sigma(3,:) - Z*resum_expanded(3,:) &
           &          - Z*(one+Z)*(two+Z)/6._dp*resum_expanded(1,:)**3/sigmatot**2 &
           &          + Z*(one+Z)/two*resum_expanded(1,:)**2/sigmatot**2*(sigma(1,:) + Z*lumi_asympt(1)) &
           &          + Z/two*(two*sigmatot*(sigma(1,:))*(lumi_asympt(2)-resum_expanded(2,:)) &
           &          + (Z-one)*lumi_asympt(1)**2*(sigma(1,:)) + two*lumi_asympt(1)*sigmatot*(sigma(2,:) - Z*resum_expanded(2,:)))/sigmatot**2 &
           &          - Z/two*resum_expanded(1,:)*(Z*lumi_asympt(1)*(two*(sigma(1,:)) + (Z-one)*lumi_asympt(1)) &
           &          + two*sigmatot*(sigma(2,:) + Z*lumi_asympt(2) - (one+Z)*resum_expanded(2,:)))/sigmatot**2
   end if

   res =  (resummation/lumi_asympt(0))**Z*matching_factor

  end function matching_mult_a_matrix

  function matching_a(sigma, pt, resummation, resum_expanded, sigmabar,dlumi_lumi,dlumi_lumi_expanded) result(res)
    real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:)
    real(dp), intent(in) :: dlumi_lumi(:), dlumi_lumi_expanded(:,:)
    real(dp)             :: res(size(pt))
    !----------------------------------------------
    integer              :: order
    real(dp)             :: matching_factor(size(resummation))

    !TODO: to be fixed (check if everything is fine)
    call wae_error("matching_a: scheme (a) not yet supported")
    stop

    order = size(resum_expanded,dim=1) - 1

    matching_factor = sigma(0)*(1+dlumi_lumi) + sigma(1)+sigmabar(1,:) -resum_expanded(1,:)

    if (order >= 2) then
       matching_factor = matching_factor &
            &          + sigma(2) + sigmabar(2,:) - resum_expanded(2,:) &
            &          - (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))&
            &                           * (sigma(1)+sigmabar(1,:) - resum_expanded(1,:))
    end if

    if (order >= 3) then
       matching_factor = matching_factor &
            &          + sigma(3) + sigmabar(3,:) - resum_expanded(3,:) &
            &          - (resum_expanded(2,:)/sigma(0)-dlumi_lumi_expanded(2,:))&
            &                           * (sigma(1)+sigmabar(1,:) - resum_expanded(1,:)) &
            &          - (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))&
            &                           * (sigma(2)+sigmabar(2,:) - resum_expanded(2,:)) &
            &          + resum_expanded(1,:)/sigma(0)*(sigma(1)+sigmabar(1,:) - resum_expanded(1,:)) &
            &                           * (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))
    end if

    res =  (resummation/sigma(0)/(1+dlumi_lumi))*matching_factor

  end function matching_a


  function matching_b(sigma, pt, resummation, resum_expanded, sigmabar,dlumi_lumi,dlumi_lumi_expanded) result(res)
    real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:)
    real(dp), intent(in) :: dlumi_lumi(:), dlumi_lumi_expanded(:,:)
    real(dp)             :: res(size(pt))
    !----------------------------------------------
    integer              :: order
    real(dp)             :: matching_factor(size(resummation))

    order = size(resum_expanded,dim=1) - 1

    matching_factor = sigma(0)*(1+dlumi_lumi) +sigmabar(1,:) - resum_expanded(1,:)

    if (order >= 2) then
       matching_factor = matching_factor + sigma(1) + sigmabar(2,:) - resum_expanded(2,:)&
            &          - (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))&
            &          * (sigma(1)+sigmabar(1,:) - resum_expanded(1,:))
    end if

    if (order >= 3) then
       matching_factor = matching_factor &
            &          + sigma(2) + sigmabar(3,:) - resum_expanded(3,:) &
            &          - (resum_expanded(2,:)/sigma(0)-dlumi_lumi_expanded(2,:))&
            &                           * (sigma(1)+sigmabar(1,:) - resum_expanded(1,:)) &
            &          - (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))&
            &                           * (sigma(2)+sigmabar(2,:) - resum_expanded(2,:)) &
            &          + resum_expanded(1,:)/sigma(0)*(sigma(1)+sigmabar(1,:) - resum_expanded(1,:)) &
            &                           * (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))
    end if

    res =  (resummation/sigma(0)/(1+dlumi_lumi))*matching_factor

  end function matching_b


  function matching_c(sigma, pt, resummation, resum_expanded, sigmabar,dlumi_lumi,dlumi_lumi_expanded) result(res)
    real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:)
    real(dp), intent(in) :: dlumi_lumi(:), dlumi_lumi_expanded(:,:)
    real(dp)             :: res(size(pt))
    !----------------------------------------------
    integer              :: order
    real(dp)             :: matching_factor(size(resummation))

    order = size(resum_expanded,dim=1) - 1

    if (order < 2) call wae_error("matching_c: scheme (c) not supported for fixed order < 2")

    if (order >= 2) then
       matching_factor = sigma(0)*(1+dlumi_lumi) + sigmabar(1,:) - resum_expanded(1,:) &
            &          + sigmabar(2,:) - resum_expanded(2,:) - sigma(1)/sigma(0)*sigmabar(1,:)&
            &          - (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))&
            &          * (sigmabar(1,:) - resum_expanded(1,:))
    end if

    if (order >= 3) then
       matching_factor = matching_factor &
            &          + sigma(1) + sigmabar(3,:) - resum_expanded(3,:) &
            &          + sigma(1)/sigma(0)*sigmabar(1,:) - sigma(2)/sigma(0)*sigmabar(1,:) &
            &          - (resum_expanded(2,:)/sigma(0)-dlumi_lumi_expanded(2,:))&
            &                           * (sigma(1)+sigmabar(1,:) - resum_expanded(1,:)) &
            &          - (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))*sigma(1) &
            &          - (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))&
            &                           * (sigmabar(2,:) - resum_expanded(2,:)) &
            &          + resum_expanded(1,:)/sigma(0)*(sigma(1)+sigmabar(1,:) - resum_expanded(1,:)) &
            &                           * (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))
    end if

    res =  (resummation/sigma(0)/(1+dlumi_lumi))*matching_factor

  end function matching_c




  function matching_d(sigma, pt, resummation, resum_expanded, sigmabar,dlumi_lumi,dlumi_lumi_expanded) result(res)
    real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:)
    real(dp), intent(in) :: dlumi_lumi(:), dlumi_lumi_expanded(:,:)
    real(dp)             :: res(size(pt))
    !----------------------------------------------
    integer              :: order
    real(dp)             :: matching_factor(size(resummation))

    order = size(resum_expanded,dim=1) - 1

    if (order < 3) call wae_error("matching_d: scheme (d) not supported for fixed order < 3")

    if (order >= 3) then
       matching_factor = sigma(0)*(1+dlumi_lumi) + sigmabar(1,:) - resum_expanded(1,:) &
            &          + sigmabar(2,:) - resum_expanded(2,:) - sigma(1)/sigma(0)*sigmabar(1,:)&
            &          - (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))&
            &          * (sigmabar(1,:) - resum_expanded(1,:)) &
            &          + sigmabar(3,:) - resum_expanded(3,:) - sigma(1)/sigma(0)*sigmabar(2,:) &
            &          + (sigma(1)**2/sigma(0)**2 - sigma(2)/sigma(0))*sigmabar(1,:) &
            &          - (resum_expanded(2,:)/sigma(0)-dlumi_lumi_expanded(2,:))&
            &                           * (sigmabar(1,:) - resum_expanded(1,:)) &
            &          - (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))&
            &                           * (sigmabar(2,:) - resum_expanded(2,:) - sigma(1)/sigma(0)*sigmabar(1,:)) &
            &          + resum_expanded(1,:)/sigma(0)*(sigmabar(1,:) - resum_expanded(1,:)) &
            &                           * (resum_expanded(1,:)/sigma(0)-dlumi_lumi_expanded(1,:))
    end if

    res =  (resummation/sigma(0)/(1+dlumi_lumi))*matching_factor

  end function matching_d


  function matching_R(sigma, pt, resummation, resum_expanded, sigmabar) result(res)
    real(dp), intent(in) :: sigma(0:), pt(:), resummation(:), resum_expanded(0:,:), sigmabar(1:,:)
    real(dp)             :: res(size(pt))
    !----------------------------------------------
    real(dp) :: Z(size(pt)), cutoff
    integer  :: i, order

    order = size(resum_expanded,dim=1) - 1
    ! pt cutoff at which we switch off the finite remainder (it can be set to zero)
    !cutoff = 2.0_dp
    cutoff = zero


    ! This matching is devised for ptB at NNLL, where all the O(as^2)
    ! constant terms are included in the resummation
    res = resummation

    do i = 1, size(pt)
       if (pt(i) .ge. cutoff) then
          res(i) = res(i) + sigma(1)+sigmabar(1,i)-resum_expanded(1,i)
          if (order .ge. 2) res(i) = res(i) + sigma(2)+sigmabar(2,i)&
               & -resum_expanded(2,i)
          if (order .eq. 3) res(i) = res(i) + sigma(3)+sigmabar(3,i)&
               & -resum_expanded(3,i)
       end if
    end do


  end function matching_R

  function matching_R_matrix(sigmatot, pt, resummation, resum_expanded, sigma) result(res)
   real(dp), intent(in) :: sigmatot, pt(:), resummation(:), resum_expanded(0:,:), sigma(1:,:)
   real(dp)             :: res(size(pt))
   !----------------------------------------------
   real(dp) :: Z(size(pt)), cutoff
   integer  :: i, order

   order = size(resum_expanded,dim=1) - 1
   ! pt cutoff at which we switch off the finite remainder (it can be set to zero)
   !cutoff = 2.0_dp
   cutoff = zero


   ! This matching is devised for ptB at NNLL, where all the O(as^2)
   ! constant terms are included in the resummation
   res = resummation

   do i = 1, size(pt)
      if (pt(i) .ge. cutoff) then
         res(i) = res(i) + sigma(1,i)-resum_expanded(1,i)
         if (order .ge. 2) res(i) = res(i) + sigma(2,i)&
              & -resum_expanded(2,i)
         if (order .eq. 3) res(i) = res(i) + sigma(3,i)&
              & -resum_expanded(3,i)
      end if
   end do


 end function matching_R_matrix

  !----------------------------------------------------------------------
  ! Function that returns the result in a given fixed-order scheme;
  ! the order to be used is deduced from the number of columns in the
  ! sigmabar array.
  !
  ! Input parameters:
  !   - scheme = string indicating the name of the scheme
  !   - sigma(0:fixed_order) = terms in the total cross section from LO(0) up to fixed_order
  !   - sigmabar(1:fixed_order, :) = 1-jet cross section, integrated from infinity down to pt (2nd index)
  !
  function fixed_order_schemes(scheme, sigma, sigmabar) result(res)
    character(len=*), intent(in) :: scheme
    real(dp), intent(in) :: sigma(0:), sigmabar(1:,:)
    real(dp)             :: res(size(sigmabar,dim=2)), norm
    integer              :: fixed_order

    fixed_order = size(sigmabar,dim=1)
    if (size(sigma)-1 < fixed_order) call wae_error("fixed_order_schemes", &
         & "size of sigma(:) array is smaller than fixed_order deduced from sigmabar")

    select case(trim(scheme))
    case("a")
       norm = sum(sigma(:fixed_order))
       res  = one + sum(sigmabar,dim=1)/norm
    case("b")
       norm = sum(sigma(0:fixed_order-1))
       res  = one + sum(sigmabar,dim=1)/norm
    case("c")
       if (fixed_order < 2) call wae_error("fixed_order_schemes: scheme (c) unsupported for fixed_order < 2")
       norm = sum(sigma(0:fixed_order-2))
       res = one + sum(sigmabar,dim=1)/norm &
            &      - sigma(fixed_order-1)/norm**2*sigmabar(1,:)
    case("d")
       ! check out the following expression
       if (fixed_order < 3) call wae_error("fixed_order_schemes: scheme (d) unsupported for fixed_order < 3")
       norm = sum(sigma(0:fixed_order-3))
       res = one + sum(sigmabar,dim=1)/norm &
            &      - sigma(fixed_order-2)/norm**2*(sigmabar(1,:)+sigmabar(2,:)) &
            &      + (sigma(1)*sigma(fixed_order-2)-sigma(0)*sigma(fixed_order-1))*sigmabar(1,:)/norm**3
!       res = one + sum(sigmabar,dim=1)/norm &
!            &      - sigma(fixed_order-1)/norm**2*(sigmabar(1,:)+sigmabar(2,:)) &
!            &      + (sigma(1)*sigma(fixed_order-1)-sigma(0)*sigma(fixed_order-1))*sigmabar(1,:)/norm**3
    case("moda")
       norm = sum(sigma(:fixed_order))
       res  = one + sum(sigmabar,dim=1)/norm
    case("modRa")
       norm = sum(sigma(:fixed_order))
       res  = one + sum(sigmabar,dim=1)/norm
    case("R")
       norm = sum(sigma(:fixed_order))
       res  = one + sum(sigmabar,dim=1)/norm
    case("loga")
       norm = sum(sigma(:fixed_order))
       res  = one + sum(sigmabar,dim=1)/norm
    case("mult")
       norm = sum(sigma(:fixed_order))
       res  = one + sum(sigmabar,dim=1)/norm
    case("multa")
       norm = sum(sigma(:fixed_order))
       res  = one + sum(sigmabar,dim=1)/norm
    case default
       call wae_error("fixed_order_schemes: unrecognised fixed order scheme "//scheme)
    end select
  end function fixed_order_schemes


end module matching
